"""Integration for associating review requests with Asana tasks."""

from __future__ import annotations

from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from reviewboard.extensions.hooks import ReviewRequestFieldsHook
from reviewboard.integrations.base import Integration

from rbintegrations.asana.fields import AsanaField
from rbintegrations.asana.forms import AsanaIntegrationConfigForm


class AsanaIntegration(Integration):
    """Integrates Review Board with Asana."""

    name = 'Asana'
    description = _('Associate Asana tasks with your review requests.')
    config_form_cls = AsanaIntegrationConfigForm

    def initialize(self) -> None:
        """Initialize the integration hooks."""
        ReviewRequestFieldsHook(self, 'main', [AsanaField])

    @cached_property
    def icon_static_urls(self) -> dict[str, str]:
        """The icons used for the integration."""
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        return {
            '1x': extension.get_static_url('images/asana/icon.png'),
            '2x': extension.get_static_url('images/asana/icon@2x.png'),
        }



"""Views for the Travis CI integration (webhook listener)."""

import base64
import json
import logging

import cryptography
from django.http import HttpResponse, HttpResponseBadRequest
from django.views.generic import View
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.models.status_update import StatusUpdate

from rbintegrations.travisci.api import TravisAPI


logger = logging.getLogger(__name__)


class TravisCIWebHookView(View):
    """The view to handle webhook notifications from a Travis CI build."""

    def post(self, request, *args, **kwargs):
        """Handle the POST.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Additional positional arguments, parsed from the URL.

            **kwargs (dict):
                Additional keyword arguments, parsed from the URL.

        Returns:
            django.http.HttpResponse:
            A response.
        """
        payload = json.loads(request.POST['payload'])

        try:
            payload_env = payload['matrix'][0]['config']['env']
        except (IndexError, KeyError):
            error = 'Got event without an env in config.'
            logger.error('Travis CI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        integration_config_id = None
        status_update_id = None

        for line in payload_env:
            key, value = line.split('=', 1)

            if key == 'REVIEWBOARD_STATUS_UPDATE_ID':
                status_update_id = int(value)
            elif key == 'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID':
                integration_config_id = int(value)

        if status_update_id is None:
            error = 'Unable to find REVIEWBOARD_STATUS_UPDATE_ID in payload.'
            logger.error('Travis CI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        if integration_config_id is None:
            error = ('Unable to find REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID '
                     'in payload.')
            logger.error('Travis CI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        logger.debug('Got Travis CI webhook event for Integration Config %d '
                     '(status update %d)',
                     integration_config_id, status_update_id,
                     extra={'request': request})

        try:
            integration_config = IntegrationConfig.objects.get(
                pk=integration_config_id)
        except IntegrationConfig.DoesNotExist:
            error = ('Unable to find matching integration config ID %d.'
                     % integration_config_id)
            logger.error('Travis CI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        if not self._validate_signature(request, integration_config):
            error = ('Invalid Travis CI webhook signature for status '
                     'update %d.'
                     % status_update_id)
            return HttpResponseBadRequest(error)

        try:
            status_update = StatusUpdate.objects.get(pk=status_update_id)
        except StatusUpdate.DoesNotExist:
            error = ('Unable to find matching status update ID %d.'
                     % status_update_id)
            logger.error('Travis CI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        status_update.url = payload['build_url']
        status_update.url_text = 'View Build'

        build_state = payload['state']

        if build_state == 'passed':
            status_update.state = StatusUpdate.DONE_SUCCESS
            status_update.description = 'build succeeded.'
        elif build_state == 'started':
            status_update.state = StatusUpdate.PENDING
            status_update.description = 'building...'
        elif build_state in ('errored', 'failed'):
            status_update.state = StatusUpdate.DONE_FAILURE
            status_update.description = 'build failed.'
        else:
            logger.warning('Got unexpected Travis CI build state "%s"',
                           build_state)

        status_update.save()

        return HttpResponse()

    def _validate_signature(self, request, integration_config):
        """Validate the webhook signature.

        This will fetch the public key from the appropriate Travis CI server
        and use it to verify the signature of the payload.

        Args:
            request (django.http.HttpRequest):
                The HTTP request for the webhook.

            integration_config (reviewboard.integrations.models.
                                IntegrationConfig):
                The integration configuration that requested the Travis CI job.

        Returns:
            bool:
            True if the signature validated correctly.
        """
        api = TravisAPI(integration_config)

        try:
            data = api.get_config()
        except Exception as e:
            logger.error('Failed to fetch config information from Travis CI '
                         'server at %s: %s',
                         api.endpoint, e,
                         extra={'request': request})
            return False

        try:
            crypto_primitives = cryptography.hazmat.primitives

            signature = base64.b64decode(request.META['HTTP_SIGNATURE'])
            backend = cryptography.hazmat.backends.default_backend()
            public_key = \
                data['config']['notifications']['webhook']['public_key']
            key = crypto_primitives.serialization.load_pem_public_key(
                public_key.encode('ascii'), backend)

            key.verify(
                signature,
                request.POST['payload'].encode('ascii'),
                crypto_primitives.asymmetric.padding.PKCS1v15(),
                crypto_primitives.hashes.SHA1())
            return True
        except cryptography.exceptions.InvalidSignature:
            logger.error('Unable to verify signature for Travis CI webhook.',
                         extra={'request': request})
            return False
        except Exception as e:
            logger.exception('Unexpected error while verifying Travis CI '
                             'signature: %s',
                             e, extra={'request': request})
            return False



"""Entries that can be posted to I Done This on behalf of the user."""

import re
import string

from reviewboard.admin.server import build_server_url


# Entry types.
REPLY_PUBLISHED = 'reply_published'
REVIEW_REQUEST_COMPLETED = 'review_request_completed'
REVIEW_REQUEST_DISCARDED = 'review_request_discarded'
REVIEW_REQUEST_PUBLISHED = 'review_request_published'
REVIEW_REQUEST_REOPENED = 'review_request_reopened'
REVIEW_REQUEST_UPDATED = 'review_request_updated'
REVIEW_PUBLISHED = 'review_published'
REVIEW_PUBLISHED_ISSUE = 'review_published_issue'
REVIEW_PUBLISHED_ISSUES = 'review_published_issues'
REVIEW_PUBLISHED_SHIPIT = 'review_published_shipit'
REVIEW_PUBLISHED_SHIPIT_ISSUE = 'review_published_shipit_issue'
REVIEW_PUBLISHED_SHIPIT_ISSUES = 'review_published_shipit_issues'


# Default template strings for each entry type. Should not be translated.
default_template_strings = {
    REPLY_PUBLISHED:
        'Replied to review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_REQUEST_COMPLETED:
        'Completed review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_REQUEST_DISCARDED:
        'Discarded review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_REQUEST_PUBLISHED:
        'Published review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_REQUEST_REOPENED:
        'Reopened review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_REQUEST_UPDATED:
        'Updated review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_PUBLISHED:
        'Posted review on review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_PUBLISHED_ISSUE:
        'Posted review (1 issue) on review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_PUBLISHED_ISSUES:
        'Posted review (${num_issues} issues) on review request '
        '${review_request_id}: ${summary} ${url} ${group_tags}',
    REVIEW_PUBLISHED_SHIPIT:
        'Posted Ship it! on review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_PUBLISHED_SHIPIT_ISSUE:
        'Posted Ship it! (1 issue) on review request ${review_request_id}: '
        '${summary} ${url} ${group_tags}',
    REVIEW_PUBLISHED_SHIPIT_ISSUES:
        'Posted Ship it! (${num_issues} issues) on review request '
        '${review_request_id}: ${summary} ${url} ${group_tags}',
}


# Tags allow only alphanumeric characters and underscore.
INVALID_TAG_CHARS_RE = re.compile(r'\W')
MULTIPLE_WHITESPACE_RE = re.compile(r'\s+')


def format_template_string(template_string, num_issues=0, review_request=None,
                           review_request_id=None, summary=None,
                           group_tags=None, url=None):
    """Format a template string for an I Done This entry.

    Args:
        template_string (unicode):
            The template string to substitute arguments into.

        num_issues (int, optional):
            Number of issues opened in a review, used for ``${num_issues}``.

        review_request (reviewboard.reviews.models.review_request.
                        ReviewRequest, optional):
            Review request used for the remaining template string arguments.
            Any arguments provided separately will override the review request.

        review_request_id (int, optional):
            Review request ID, used for ``${review_request_id}``.

        summary (unicode, optional):
            Review request summary, used for ``${summary}``.

        group_tags (unicode, optional):
            Reviewer group names as #tags, used for ``${group_tags}``.

        url (unicode, optional):
            URL for the review request, review, or reply, used for ``${url}``.

    Returns:
        unicode:
        Template string with replaced arguments and cleaned whitespace.

    Raises:
        ValueError:
            Raised if the template string is invalid.

        KeyError:
            Raised if the template string contains unrecognized arguments.
    """
    if review_request:
        review_request_id = review_request_id or review_request.display_id
        summary = summary or review_request.summary
        url = url or build_server_url(review_request.get_absolute_url())

        if not group_tags:
            # Tags allow only alphanumeric characters and underscore.
            group_names = review_request.target_groups.values_list('name',
                                                                   flat=True)
            group_tags = ' '.join(
                '#%s' % INVALID_TAG_CHARS_RE.sub('_', group_name)
                for group_name in group_names)

    result = string.Template(template_string).substitute(
        num_issues=num_issues,
        review_request_id=review_request_id,
        summary=summary,
        group_tags=group_tags,
        url=url)

    return MULTIPLE_WHITESPACE_RE.sub(' ', result).strip()



"""The form for configuring the Trello integration."""

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm

from rbintegrations.util.conditions import ReviewRequestConditionsField


class TrelloIntegrationConfigForm(IntegrationConfigForm):
    """Form for configuring Trello."""

    conditions = ReviewRequestConditionsField()

    trello_api_key = forms.CharField(
        label=_('Trello API Key'),
        help_text=_('Your Trello API key. This can be created by registering '
                    'an application in the <a '
                    'href="https://trello.com/power-ups/admin/">Trello '
                    'Power-Up Admin Portal</a>.'),
        widget=forms.widgets.TextInput(attrs={
            'size': 40,
        }))

    trello_api_token = forms.CharField(
        label=_('Trello API Token'),
        help_text=_('An access token for the Trello API. To create this, '
                    'after creating a Power-Up in the <a '
                    'href="https://trello.com/power-ups/admin/">Trello '
                    'Power-Up Admin Portal</a>, select "manually generate a '
                    'Token".'),
        widget=forms.widgets.TextInput(attrs={
            'size': 40,
        }))



"""Unit tests for the Discord integration."""

import json
from urllib.request import urlopen

from django.contrib.auth.models import User
from djblets.conditions import ConditionSet, Condition
from djblets.testing.decorators import add_fixtures
from reviewboard.accounts.trophies import TrophyType, trophies_registry
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import ReviewRequestDraft

from rbintegrations.discord.integration import DiscordIntegration
from rbintegrations.testing.testcases import IntegrationTestCase


class DiscordIntegrationTests(IntegrationTestCase):
    """Tests Review Board integration with Discord."""

    integration_cls = DiscordIntegration
    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(DiscordIntegrationTests, self).setUp()

        self.user = User.objects.create(username='test',
                                        first_name='Test',
                                        last_name='User')

    def test_notify_new_review_request(self):
        """Testing DiscordIntegration notifies on new review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    @add_fixtures(['test_site'])
    def test_notify_new_review_request_with_local_site(self):
        """Testing DiscordIntegration notifies on new review request with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            target_people=[self.user],
            publish=False)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/s/local-site-1/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/s/local-site-1/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/s/local-site-1/users/test/'
                    '|Test User>'
                ),
            })

    def test_notify_new_review_request_with_diff(self):
        """Testing DiscordIntegration notifies on new review request with
        diff
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)
        self.create_diffset(review_request)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Diff',
                            'value': (
                                '<http://example.com/r/1/diff/1/|Revision 1>'
                            ),
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    @add_fixtures(['test_site'])
    def test_notify_new_review_request_with_local_site_and_diff(self):
        """Testing DiscordIntegration notifies on new review request with local
        site and diff
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=False)
        self.create_diffset(review_request)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/s/local-site-1/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Diff',
                            'value': (
                                '<http://example.com/s/local-site-1/r/1/'
                                'diff/1/|Revision 1>'
                            ),
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/s/local-site-1/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/s/local-site-1/users/test/'
                    '|Test User>'
                ),
            })

    def test_notify_new_review_request_with_image_file_attachment(self):
        """Testing DiscordIntegration notifies on new review request with
        image file attachment
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)
        attachment = self.create_file_attachment(review_request)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'image_url': attachment.get_absolute_url(),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_request_with_fish_trophy(self):
        """Testing DiscordIntegration notifies on new review request with
        fish trophy
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False,
            id=12321)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#12321: New review request from Test User: '
                        'http://example.com/r/12321/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'thumb_url': self.integration.trophy_urls['fish'],
                    'title': '#12321: Test Review Request',
                    'title_link': 'http://example.com/r/12321/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_request_with_milestone_trophy(self):
        """Testing DiscordIntegration notifies on new review request with
        milestone trophy
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False,
            id=10000)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#10000: New review request from Test User: '
                        'http://example.com/r/10000/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'thumb_url': self.integration.trophy_urls['milestone'],
                    'title': '#10000: Test Review Request',
                    'title_link': 'http://example.com/r/10000/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_request_with_custom_trophy(self):
        """Testing DiscordIntegration notifies on new review request with
        ignored custom trophy
        """
        class MyTrophy(TrophyType):
            category = 'test'

            def __init__(self):
                super(MyTrophy, self).__init__(title='My Trophy',
                                               image_url='blahblah')

            def qualifies(self, review_request):
                return True

        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)

        trophies_registry.register(MyTrophy)

        try:
            review_request.publish(self.user)
        finally:
            trophies_registry.unregister(MyTrophy)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': 'My description.',
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_request_with_long_text(self) -> None:
        """Testing DiscordIntegration.notify with long text on a field"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='X' * 2000,
            target_people=[self.user],
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertSpyCallCount(self.integration.notify, 1)
        self.assertSpyCallCount(urlopen, 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review request from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': False,
                            'title': 'Description',
                            'value': '%s...' % ('X' * 1021),
                        },
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'New review request from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_updated_review_request(self):
        """Testing DiscordIntegration notifies on updated review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New update from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: My new summary',
                    'title_link': 'http://example.com/r/1/',
                    'text': '',
                    'pretext': None,
                }],
                'text': (
                    'New update from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_updated_review_request_with_change_description(self):
        """Testing DiscordIntegration notifies on updated review request
        with change description
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        changedesc = draft.changedesc
        changedesc.text = 'These are my changes.'
        changedesc.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New update from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'title': '#1: My new summary',
                    'title_link': 'http://example.com/r/1/',
                    'text': 'These are my changes.',
                    'pretext': None,
                }],
                'text': (
                    'New update from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_updated_review_request_with_new_image_attachments(self):
        """Testing DiscordIntegration notifies on updated review request with
        new image attachments
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)
        self.create_file_attachment(review_request)
        review_request.publish(self.user)

        attachment = self.create_file_attachment(review_request,
                                                 caption='My new attachment',
                                                 draft=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New update from Test User: '
                        'http://example.com/r/1/'
                    ),
                    'fields': [
                        {
                            'short': True,
                            'title': 'Repository',
                            'value': 'Test Repo',
                        },
                        {
                            'short': True,
                            'title': 'Branch',
                            'value': 'my-branch',
                        },
                    ],
                    'image_url': attachment.get_absolute_url(),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': '',
                    'pretext': None,
                }],
                'text': (
                    'New update from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_closed_review_request_as_submitted(self):
        """Testing DiscordIntegration notifies on closing review request as
        submitted
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.close(review_request.SUBMITTED)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: Closed as completed by Test User: '
                        'http://example.com/r/1/'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'Closed as completed by '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_closed_review_request_as_discarded(self):
        """Testing DiscordIntegration notifies on closing review request as
        discarded
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.close(review_request.DISCARDED)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: Discarded by Test User: '
                        'http://example.com/r/1/'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'Discarded by '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    @add_fixtures(['test_site'])
    def test_notify_closed_review_request_with_local_site(self):
        """Testing DiscordIntegration notifies on closing review request with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.close(review_request.SUBMITTED)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: Closed as completed by Test User: '
                        'http://example.com/s/local-site-1/r/1/'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/s/local-site-1/r/1/',
                    'text': None,
                    'pretext': None,
                }],
                'text': (
                    'Closed as completed by '
                    '<http://example.com/s/local-site-1/users/test/'
                    '|Test User>'
                ),
            })

    def test_notify_reopened_review_request(self):
        """Testing DiscordIntegration notifies on reopened review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)
        review_request.close(review_request.SUBMITTED)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.reopen(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: Reopened by Test User: '
                        'http://example.com/r/1/'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/',
                    'text': 'My description.',
                    'pretext': None
                }],
                'text': (
                    'Reopened by '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    @add_fixtures(['test_site'])
    def test_notify_reopened_review_request_with_local_site(self):
        """Testing DiscordIntegration notifies on reopened review request with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=True)
        review_request.close(review_request.SUBMITTED)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.reopen(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: Reopened by Test User: '
                        'http://example.com/s/local-site-1/r/1/'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/s/local-site-1/r/1/',
                    'text': 'My description.',
                    'pretext': None,
                }],
                'text': (
                    'Reopened by '
                    '<http://example.com/s/local-site-1/users/test/'
                    '|Test User>'
                ),
            })

    def test_notify_new_review_with_body_top(self):
        """Testing DiscordIntegration notifies on new review with body_top"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='This is my review.')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review from Test User: '
                        'http://example.com/r/1/#review1'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'This is my review.',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    @add_fixtures(['test_site'])
    def test_notify_new_review_with_local_site(self):
        """Testing DiscordIntegration notifies on new review with local site"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='This is my review.')
        self.create_general_comment(review)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review from Test User: '
                        'http://example.com/s/local-site-1/r/1/#review1'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': (
                        'http://example.com/s/local-site-1/r/1/#review1'
                    ),
                    'text': 'This is my review.',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/s/local-site-1/users/test/'
                    '|Test User>'
                ),
            })

    def test_notify_new_review_with_comments(self):
        """Testing DiscordIntegration notifies on new review with only comments
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.')

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New review from Test User: '
                        'http://example.com/r/1/#review1'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'My general comment.',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_with_one_open_issue(self):
        """Testing DiscordIntegration notifies on new review with 1 open
        issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': 'warning',
                    'fallback': (
                        '#1: New review from Test User (1 issue): '
                        'http://example.com/r/1/#review1'
                    ),
                    'fields': [{
                        'title': 'Open Issues',
                        'value': ':warning: 1 issue',
                        'short': True,
                    }],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'My general comment.',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_with_open_issues(self):
        """Testing DiscordIntegration notifies on new review with > 1 open
        issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.',
                                    issue_opened=True)
        self.create_general_comment(review, text='My general comment 2.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': 'warning',
                    'fallback': (
                        '#1: New review from Test User (2 issues): '
                        'http://example.com/r/1/#review1'
                    ),
                    'fields': [{
                        'title': 'Open Issues',
                        'value': ':warning: 2 issues',
                        'short': True,
                    }],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'My general comment.',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_with_ship_it(self):
        """Testing DiscordIntegration notifies on new review with Ship It!"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review,
                                    text='My comment')

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': 'good',
                    'fallback': (
                        '#1: New review from Test User (Ship it!): '
                        'http://example.com/r/1/#review1'
                    ),
                    'fields': [{
                        'title': 'Ship it!',
                        'value': ':white_check_mark:',
                        'short': True,
                    }],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'My comment',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_with_ship_it_and_custom_body_top(self):
        """Testing DiscordIntegration notifies on new review with Ship It
        and custom body_top
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            target_people=[self.user],
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='This is body_top.')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': 'good',
                    'fallback': (
                        '#1: New review from Test User (Ship it!): '
                        'http://example.com/r/1/#review1'
                    ),
                    'fields': [{
                        'title': 'Ship it!',
                        'value': ':white_check_mark:',
                        'short': True,
                    }],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'This is body_top.',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_with_ship_it_and_one_open_issue(self):
        """Testing DiscordIntegration notifies on new review with Ship It!
        and 1 open issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            target_people=[self.user],
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review,
                                    text='My comment',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': 'warning',
                    'fallback': (
                        '#1: New review from Test User '
                        '(Fix it, then Ship it!): '
                        'http://example.com/r/1/#review1'
                    ),
                    'fields': [{
                        'title': 'Fix it, then Ship it!',
                        'value': ':warning: 1 issue',
                        'short': True,
                    }],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'My comment',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_new_review_with_ship_it_and_open_issues(self):
        """Testing DiscordIntegration notifies on new review with Ship It!
        and > 1 open issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            target_people=[self.user],
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review,
                                    text='My comment 1',
                                    issue_opened=True)
        self.create_general_comment(review,
                                    text='My comment 2',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': 'warning',
                    'fallback': (
                        '#1: New review from Test User '
                        '(Fix it, then Ship it!): '
                        'http://example.com/r/1/#review1'
                    ),
                    'fields': [{
                        'title': 'Fix it, then Ship it!',
                        'value': ':warning: 2 issues',
                        'short': True,
                    }],
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'My comment 1',
                    'pretext': None,
                }],
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def test_notify_with_body_bottom(self):
        """Testing DiscordIntegration notifies on new review with body_bottom
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='',
                                    body_bottom='Test')

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'attachments': [{
                    'color': '#efcc96',
                    'fallback': (
                        '#1: New review from Test User: '
                        'http://example.com/r/1/#review1'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'Test',
                    'pretext': None,
                }],
                'icon_url': self.integration.logo_url,
                'text': (
                    'New review from '
                    '<http://example.com/users/test/|Test User>'
                ),
                'username': 'RB User',
            })

    def test_notify_new_reply_with_body_top(self):
        """Testing DiscordIntegration notifies on new reply with body_top"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            target_people=[self.user],
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review,
                                  user=self.user,
                                  body_top='This is body_top.')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        reply.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New reply from Test User: '
                        'http://example.com/r/1/#review1'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#review1',
                    'text': 'This is body_top.',
                    'pretext': None,
                }],
                'text': (
                    'New reply from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    @add_fixtures(['test_site'])
    def test_notify_new_reply_with_local_site(self):
        """Testing DiscordIntegration notifies on new reply with local
        site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            target_people=[self.user],
            publish=True)

        review_request.local_site.users.add(self.user)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review,
                                  user=self.user,
                                  body_top='This is body_top.')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        reply.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New reply from Test User: '
                        'http://example.com/s/local-site-1/r/1/#review1'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': (
                        'http://example.com/s/local-site-1/r/1/#review1'
                    ),
                    'text': 'This is body_top.',
                    'pretext': None,
                }],
                'text': (
                    'New reply from '
                    '<http://example.com/s/local-site-1/users/test/'
                    '|Test User>'
                ),
            })

    def test_notify_new_reply_with_comment(self):
        """Testing DiscordIntegration notifies on new reply with comment"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            target_people=[self.user],
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review, user=self.user, body_top='')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        reply.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'username': 'RB User',
                'icon_url': self.integration.logo_url,
                'attachments': [{
                    'color': self.integration.default_color,
                    'fallback': (
                        '#1: New reply from Test User: '
                        'http://example.com/r/1/#gcomment2'
                    ),
                    'title': '#1: Test Review Request',
                    'title_link': 'http://example.com/r/1/#gcomment2',
                    'text': 'This is a comment.',
                    'pretext': None,
                }],
                'text': (
                    'New reply from '
                    '<http://example.com/users/test/|Test User>'
                ),
            })

    def _create_config(self, with_local_site=False):
        """Set configuration values for DiscordIntegration"""
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any')),
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config 1',
                                                enabled=True,
                                                local_site=local_site)
        config.set('notify_username', 'RB User')
        config.set('webhook_url', 'http://example.com/discord-url/')
        config.set('conditions', condition_set.serialize())
        config.save()

        return config



"""The form for configuring the Asana integration."""

from __future__ import annotations

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm

from rbintegrations.util.conditions import ReviewRequestConditionsField


class AsanaIntegrationConfigForm(IntegrationConfigForm):
    """Form for configuring Asana."""

    conditions = ReviewRequestConditionsField()

    asana_access_token = forms.CharField(
        label=_('Asana Personal Access Token'),
        help_text=(
            _('A <a href="%s">personal access token</a> for authenticating to '
              'the Asana API.')
            % 'https://developers.asana.com/docs/personal-access-token'),
        min_length=34,
        max_length=128,
        widget=forms.widgets.TextInput(attrs={
            'size': 128,
        }))

    asana_workspace = forms.ChoiceField(label=_('Asana Workspace'),
                                        choices=[])

    asana_workspace_name = forms.CharField(widget=forms.HiddenInput())

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the form.

        Args:
            *args (tuple):
                Positional arguments for the form.

            **kwargs (dict):
                Keyword arguments for the form.
        """
        super().__init__(*args, **kwargs)

        from rbintegrations.extension import RBIntegrationsExtension
        extension = RBIntegrationsExtension.instance
        assert extension is not None

        asana_integration_config_bundle = \
            extension.get_bundle_id('asana-integration-config')
        self.css_bundle_names = [asana_integration_config_bundle]
        self.js_bundle_names = [asana_integration_config_bundle]

        if (self.data and
            'asana_workspace' in self.data and
            'asana_workspace_name' in self.data):
            # Saving case: If the provided form data contains the workspace
            # ID and name, set those as valid choices so that the
            # asana_workspace field will validate properly.

            self.fields['asana_workspace'].choices = [
                (self.data['asana_workspace'],
                 self.data['asana_workspace_name']),
            ]
        elif (self.instance and
              'asana_workspace' in self.instance.settings and
              'asana_workspace_name' in self.instance.settings):
            # Loading case: If the saved settings contain the workspace ID and
            # name, set those as valid choices for the asana_workspace field so
            # that it shows the initial value correctly.

            self.fields['asana_workspace'].choices = [
                (self.instance.settings['asana_workspace'],
                 self.instance.settings['asana_workspace_name']),
            ]



"""Form-related utilities.

Version Added:
    4.1
"""

from __future__ import annotations

from typing import TYPE_CHECKING


try:
    # Review Board >= 7.1
    from reviewboard.reviews.conditions import review_request_condition_choices
    from reviewboard.reviews.forms import ReviewRequestConditionsField
except ImportError:
    # Review Board < 7.1
    if TYPE_CHECKING:
        assert False
        from djblets.conditions.choices import ConditionChoices
    else:
        from django.utils.translation import gettext_lazy as _
        from djblets.forms.fields import ConditionsField
        from djblets.util.typing import StrOrPromise
        from reviewboard.reviews.conditions import \
            ReviewRequestConditionChoices


    review_request_condition_choices = \
        ReviewRequestConditionChoices()  # noqa: E303

    class ReviewRequestConditionsField(ConditionsField):
        """Compatibility field for selecting review request conditions.

        This can be removed once Review Board 7.1 or higher is required.

        Version Added:
            4.1
        """

        def __init__(
            self,
            *,
            choices: ConditionChoices = review_request_condition_choices,
            label: StrOrPromise = _('Conditions'),
            **kwargs,
        ) -> None:
            """Initialize the field.

            Args:
                choices (djblets.conditions.choices.ConditionChoices or type or
                         callable):
                    The registry of choices available for the conditions.

                    This defaults to the standard review request condition
                    choices. It can be overridden to provide a more
                    specialized list of conditions.

                label (str, optional):
                    A custom label to use for this field.

                **kwargs (dict):
                    Keyword arguments to pass to the parent class.
            """
            super().__init__(choices=review_request_condition_choices,
                             label=label,
                             **kwargs)


__all__ = (
    'ReviewRequestConditionsField',
    'review_request_condition_choices',
)



"""Integration for chat integrations"""

from __future__ import annotations

import logging
from typing import (ClassVar, Optional, Sequence, TYPE_CHECKING, Tuple,
                    TypedDict)

from django.utils.functional import cached_property
from djblets.db.query import get_object_or_none
from djblets.util.templatetags.djblets_utils import user_displayname
from housekeeping.functions import deprecate_non_keyword_only_args
from reviewboard.accounts.models import Trophy
from reviewboard.admin.server import build_server_url
from reviewboard.extensions.hooks import SignalHook
from reviewboard.integrations.base import Integration
from reviewboard.reviews.models import (BaseComment, Comment,
                                        FileAttachmentComment,
                                        GeneralComment,
                                        ReviewRequest,
                                        ScreenshotComment)
from reviewboard.reviews.signals import (review_request_closed,
                                         review_request_published,
                                         review_request_reopened,
                                         review_published,
                                         reply_published)
from reviewboard.site.urlresolvers import local_site_reverse

from rbintegrations.basechat.forms import BaseChatIntegrationConfigForm
from rbintegrations.deprecation import RemovedInRBIntegrations50Warning

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from djblets.integrations.forms import IntegrationConfigForm
    from reviewboard.changedescs.models import ChangeDescription
    from reviewboard.reviews.models import Review
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


class FieldsDict(TypedDict):
    """The fields comprising the information to display in a chat message.

    Version Added:
        4.0
    """

    #: Whether the field is short.
    #:
    #: This should be ``True`` if the field is short enough to be displayed
    #: side by side with other fields.
    #:
    #: This may not be relevant for some chat integrations.
    short: Optional[bool]

    #: The title for the field.
    title: str

    #: The value for the field.
    value: str


class BaseChatIntegration(Integration):
    """Integrates Review Board with chat applications.

    This will handle updating chat channels when review requests are posted,
    changed, or closed, and when there's new activity on the review request.
    """

    #: The base URL to use for static assets to use in chat messages.
    #:
    #: The static assets will usually just be the Review Board logo and
    #: trophy icons.
    #:
    #: Type:
    #:     str
    assets_base_url: ClassVar[str] = (
        'https://static.reviewboard.org/integration-assets/chat-common'
    )

    #: The timestamp to use for static assets.
    #:
    #: Type:
    #:     str
    assets_timestamp: ClassVar[str] = '?20240501-1234'

    #: The default color to style the chat message.
    #:
    #: Type:
    #:     str
    default_color: ClassVar[str] = '#efcc96'

    #: Whether to use shortcodes to represent emojis instead of unicode blocks.
    #:
    #: Version Added:
    #:     4.0
    #:
    #: Type:
    #:     bool
    use_emoji_shortcode: ClassVar[bool] = True

    #: Extensions of images that can appear in chat messages.
    #:
    #: Type:
    #:     tuple of str
    valid_image_url_exts: ClassVar[Tuple[str, ...]] = (
        '.png',
        '.bmp',
        '.gif',
        '.jpg',
        '.jpeg',
    )

    #: The form class for handling integration configuration.
    #:
    #: Type:
    #:     djblets.integrations.forms.IntegrationConfigForm
    config_form_cls: ClassVar[type[IntegrationConfigForm]] = \
        BaseChatIntegrationConfigForm

    @cached_property
    def logo_url(self) -> str:
        """The URL of the Review Board logo to use in chat messages.

        Version Changed:
            4.0:
            This is now a read-only property. Subclasses that want to
            override this must override the property instead of the
            attribute.
        """
        return f'{self.assets_base_url}/reviewboard.png{self.assets_timestamp}'

    @cached_property
    def trophy_urls(self) -> dict[str, str]:
        """URLs to the trophy images to use in chat messages.

        Version Changed:
            4.0:
            This is now a read-only property. Subclasses that want to
            override this must override the property instead of the
            attribute.
        """
        assets_base_url = self.assets_base_url
        assets_timestamp = self.assets_timestamp

        return {
            'fish': f'{assets_base_url}/fish-trophy.png{assets_timestamp}',
            'milestone':
                f'{assets_base_url}/milestone-trophy.png{assets_timestamp}',
        }

    def initialize(self) -> None:
        """Initialize the integration hooks."""
        hooks = (
            (review_request_closed, self._on_review_request_closed),
            (review_request_published, self._on_review_request_published),
            (review_request_reopened, self._on_review_request_reopened),
            (review_published, self._on_review_published),
            (reply_published, self._on_reply_published),
        )

        for signal, handler in hooks:
            SignalHook(self, signal, handler)

    @deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
    def notify(
        self,
        *,
        title: str,
        title_link: str,
        fallback_text: str,
        local_site: Optional[LocalSite],
        review_request: ReviewRequest,
        event_name: Optional[str] = None,
        fields: Sequence[FieldsDict] = [],
        pre_text: Optional[str] = None,
        body: Optional[str] = None,
        color: Optional[str] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Send a WebHook notification to chat application.

        This will post the given message to any channels configured to
        receive it.

        Version Changed:
            4.0:
            * Made all arguments keyword-only.

        Args:
            title (str):
                The title for the message.

            title_link (str):
                The link for the title of the message.

            fallback_text (str):
                The non-rich fallback text to display in the chat, for use in
                IRC and other services.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site for the review request or review emitting
                the message. Only integration configurations matching this
                Local Site will be processed.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the notification is bound to.

            event_name (str, optional):
                The name of the event triggering this notification.

            fields (list of rbintegrations.basechat.integration.FieldsDict,
                    optional):
                The fields comprising the rich message to display in chat.

            pre_text (str, optional):
                Text to display before the rest of the message.

            body (str, optional):
                The body of the message.

            color (str, optional):
                A color string or RGB hex value for the message.

            thumb_url (str, optional):
                URL of an image to show on the side of the message.

            image_url (str, optional):
                URL of an image to show in the message.
        """
        raise NotImplementedError(
            '%s must implement notify' % type(self).__name__)

    @deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
    def notify_review_or_reply(
        self,
        *,
        user: User,
        review: Review,
        pre_text: Optional[str] = None,
        fallback_text: Optional[str] = None,
        event_name: Optional[str] = None,
        first_comment: Optional[BaseComment] = None,
        **kwargs,
    ) -> None:
        """Notify chat application for any new posted reviews or replies.

        This performs the common work of notifying configured channels
        when there's a review or a reply.

        Version Changed:
            4.0:
            * Made all arguments keyword-only.

        Args:
            user (django.contrib.auth.models.User):
                The user who posted the review or reply.

            review (reviewboard.reviews.models.Review):
                The review or reply that was posted.

            pre_text (str, optional):
                Text to show before the message attachments.

            fallback_text (str, optional):
                Text to show in the fallback text, before the review URL and
                after the review request ID.

            event_name (str, optional):
                The name of the event triggering this notification.

            first_comment (reviewboard.reviews.models.BaseComment, optional):
                The first comment in a review, to generate the body message
                from. This is optional, and will be computed if needed.

            **kwargs (dict):
                Other keyword arguments to pass to :py:meth:`notify`.
        """
        review_request = review.review_request
        body = ''
        review_body_top = review.body_top

        # Prefer showing the body, unless it's just a repeat of a "Ship It!".
        # Otherwise, we'll aim for a comment, or body_bottom.
        if (review_body_top and
            (not review.ship_it or review_body_top != 'Ship It!')):
            body = review_body_top
        else:
            if not first_comment:
                for comment_cls in (Comment, FileAttachmentComment,
                                    ScreenshotComment, GeneralComment):
                    try:
                        first_comment = (
                            comment_cls.objects
                            .filter(review=review)
                            .only('text')
                        )[0]
                        break
                    except IndexError:
                        pass

            if first_comment:
                body = first_comment.text
            else:
                body = review.body_bottom

        if review.base_reply_to:
            # This is a reply. We can't use that reply's absolute URL because
            # there's no actual anchor added for that on the page. If we have a
            # comment (which we're using for the body text), we'll link to
            # that, and if not, link to the parent review.
            if first_comment:
                review_url = first_comment.get_review_url()
            else:
                review_url = review.base_reply_to.get_absolute_url()
        else:
            # This is a top-level review. We can link directly to it.
            review_url = review.get_absolute_url()

        review_url = build_server_url(review_url)

        fallback_text = '#%s: %s: %s' % (review_request.display_id,
                                         fallback_text, review_url)

        self.notify(title=self.get_review_request_title(review_request),
                    title_link=review_url,
                    fallback_text=fallback_text,
                    pre_text=pre_text,
                    body=body,
                    local_site=review.review_request.local_site,
                    review_request=review_request,
                    event_name=event_name,
                    **kwargs)

    @deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
    def notify_review_request(
        self,
        review_request: ReviewRequest,
        *,
        fallback_text: Optional[str],
        event_name: Optional[str],
        **kwargs,
    ) -> None:
        """Notify chat application for a review request update.

        This performs the common work of notifying configured channels
        when there's a new review request or update to a review request.

        Version Changed:
            4.0:
            * Made all arguments other than ``review_request`` keyword-only.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request.

            fallback_text (str):
                Text to show in the fallback text, before the review URL and
                after the review request ID.

            event_name (str):
                The name of the event triggering this notification.

            **kwargs (dict):
                Other keyword arguments to pass to :py:meth:`notify`.
        """
        review_request_url = self.get_review_request_url(review_request)
        fallback_text = '#%s: %s: %s' % (review_request.display_id,
                                         fallback_text,
                                         review_request_url)

        self.notify(title=self.get_review_request_title(review_request),
                    title_link=review_request_url,
                    fallback_text=fallback_text,
                    review_request=review_request,
                    event_name=event_name,
                    **kwargs)

    @deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
    def format_link(
        self,
        *,
        path: str,
        text: str,
    ) -> str:
        """Format the given URL and text to be shown in a Slack message.

        This will combine together the parts of the URL (method, domain, path)
        and format it using Slack's URL syntax.

        Version Changed:
            4.0:
            * Made all arguments keyword-only.

        Args:
            path (str):
                The path on the Review Board server.

            text (str):
                The text for the link.

        Returns:
            str:
            The link for use in Slack.
        """
        raise NotImplementedError(
            '%s must implement format_link' % type(self).__name__)

    def format_field_text(
        self,
        text: str,
    ) -> str:
        """Format the field text, providing any normalization required.

        This can be used by subclasses to make any changes before sending the
        text off to the Slack-compatible endpoint.

        Version Added:
            4.0

        Args:
            text (str):
                The text for the field.

        Returns:
            str:
            The formatted or normalized text.
        """
        return text

    @deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
    def get_user_text_url(
        self,
        *,
        user: User,
        local_site: Optional[LocalSite],
    ) -> str:
        """Return the URL to a user page.

        Version Changed:
            4.0:
            * Made all arguments keyword-only.

        Args:
            user (django.contrib.auth.models.User):
                The user being linked to.

            local_site (reviewboard.site.models.LocalSite):
                The local site for the link, if any.

        Returns:
            str:
            The URL to the user page.
        """
        # This doesn't use user.get_absolute_url because that won't include
        # site roots or local site names.
        return local_site_reverse(
            'user',
            local_site=local_site,
            kwargs={'username': user.username})

    @deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
    def get_user_text_link(
        self,
        *,
        user: User,
        local_site: Optional[LocalSite],
    ) -> str:
        """Return the chat application-formatted link to a user page.

        Version Changed:
            4.0:
            * Made all arguments keyword-only.

        Args:
            user (django.contrib.auth.models.User):
                The user being linked to.

            local_site (reviewboard.site.models.LocalSite):
                The local site for the link, if any.

        Returns:
            str:
            The formatted link to the user page.
        """
        return self.format_link(
            path=self.get_user_text_url(user=user, local_site=local_site),
            text=user.get_full_name() or user.username)

    def get_review_request_title(
        self,
        review_request: ReviewRequest,
    ) -> str:
        """Return the title for a review request message.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request.

        Returns:
            str:
            The title for the message.
        """
        return f'#{review_request.display_id}: {review_request.summary}'

    def get_review_request_text_link(
        self,
        review_request: ReviewRequest,
    ) -> str:
        """Return the chat application-formatted link to a review request.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being linked to.

        Returns:
            str:
            The formatted link to the review request.
        """
        return self.format_link(
            path=review_request.get_absolute_url(),
            text=review_request.summary)

    def get_review_request_url(
        self,
        review_request: ReviewRequest,
    ) -> str:
        """Return the absolute URL to a review request.

        Returns:
            str:
            The absolute URL to the review request.
        """
        return build_server_url(review_request.get_absolute_url())

    def _on_review_request_closed(
        self,
        user: User,
        review_request: ReviewRequest,
        close_type: str,
        description: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Handler for when review requests are closed.

        This will send a notification to any configured channels when
        a review request is closed.

        Args:
            user (django.contrib.auth.models.User):
                The user who closed the review request.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was closed.

            close_type (str):
                The close type.

            description (str, optional):
                The close message,

            **kwargs (dict, unused):
                Additional keyword arguments passed to the handler.
        """
        if not user:
            user = review_request.submitter

        user_link = self.get_user_text_link(
            user=user,
            local_site=review_request.local_site)

        if close_type == ReviewRequest.DISCARDED:
            pre_text = f'Discarded by {user_link}'
            fallback_text = f'Discarded by {user_displayname(user)}'
        elif close_type == ReviewRequest.SUBMITTED:
            pre_text = f'Closed as completed by {user_link}'
            fallback_text = f'Closed as completed by {user_displayname(user)}'
        else:
            logger.error('Tried to notify on review_request_closed for '
                         'review request pk=%d with unknown close type "%s"',
                         review_request.pk, close_type)
            return

        self.notify_review_request(review_request,
                                   fallback_text=fallback_text,
                                   body=description,
                                   pre_text=pre_text,
                                   local_site=review_request.local_site,
                                   event_name='review_request_closed')

    def _on_review_request_published(
        self,
        user: User,
        review_request: ReviewRequest,
        changedesc: ChangeDescription,
        **kwargs,
    ) -> None:
        """Handler for when review requests are published.

        This will send a notification to any configured channels when
        a review request is published.

        Args:
            user (django.contrib.auth.models.User):
                The user who published the review request.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was published.

            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description for the update, if any.

            **kwargs (dict, unused):
                Additional keyword arguments passed to the handler.
        """
        user_link = self.get_user_text_link(
            user=user,
            local_site=review_request.local_site)
        fields: list[FieldsDict] = []

        if changedesc:
            fallback_text = f'New update from {user_displayname(user)}'
            pre_text = f'New update from {user_link}'

            # This might be empty, which is fine. We won't show an update
            # at that point.
            body = changedesc.text
        else:
            fallback_text = f'New review request from {user_displayname(user)}'
            pre_text = f'New review request from {user_link}'
            body = None

            fields.append({
                'short': False,
                'title': 'Description',
                'value': self.format_field_text(review_request.description),
            })

        # Link to the diff in the update, if any.
        diffset = review_request.get_latest_diffset()

        if diffset:
            diff_url = local_site_reverse(
                'view-diff-revision',
                local_site=review_request.local_site,
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                })

            fields.append({
                'short': True,
                'title': 'Diff',
                'value': self.format_field_text(
                    self.format_link(path=diff_url,
                                     text=f'Revision {diffset.revision}')),
            })

        if review_request.repository:
            fields.append({
                'short': True,
                'title': 'Repository',
                'value': self.format_field_text(
                    review_request.repository.name),
            })

        if review_request.branch:
            fields.append({
                'short': True,
                'title': 'Branch',
                'value': self.format_field_text(review_request.branch),
            })

        # See if there are any new interesting file attachments to show.
        # These will only show up if the file is accessible.
        attachment = None
        valid_image_url_exts = self.valid_image_url_exts

        if changedesc:
            # Only show new files added in this change.
            try:
                new_files_pks = [
                    file[2]
                    for file in changedesc.fields_changed['files']['added']
                    if len(file) >= 3
                ]
            except KeyError:
                new_files_pks = []

            file_attachments = review_request.file_attachments.filter(
                pk__in=new_files_pks)
        else:
            # This is a new review request, so show any valid image.
            file_attachments = review_request.file_attachments.all()

        for attachment in file_attachments:
            if attachment.filename.endswith(valid_image_url_exts):
                # This one wins. Show it.
                break
        else:
            attachment = None

        if attachment:
            image_url = attachment.get_absolute_url()
        else:
            image_url = None

        # Find any trophies we may want to show in the update.
        trophies = Trophy.objects.get_trophies(review_request)
        trophy_url = None

        if trophies:
            # For now, due to the need to look up resources from a stable
            # location, we're only supporting certain trophies. First one
            # wins.
            for trophy in trophies:
                try:
                    trophy_url = self.trophy_urls[trophy.category]
                    break
                except KeyError:
                    pass

        self.notify_review_request(review_request,
                                   fallback_text=fallback_text,
                                   body=body,
                                   pre_text=pre_text,
                                   fields=fields,
                                   thumb_url=trophy_url,
                                   image_url=image_url,
                                   local_site=review_request.local_site,
                                   event_name='review_request_published')

    def _on_review_request_reopened(
        self,
        user: User,
        review_request: ReviewRequest,
        **kwargs,
    ) -> None:
        """Handler for when review requests are reopened.

        This will send a notification to any configured channels when
        a review request is reopened.

        Args:
            user (django.contrib.auth.models.User):
                The user who reopened the review request.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was published.

            **kwargs (dict, unused):
                Additional keyword arguments passed to the handler.
        """
        if not user:
            user = review_request.submitter

        user_link = self.get_user_text_link(
            user=user,
            local_site=review_request.local_site)
        pre_text = f'Reopened by {user_link}'
        fallback_text = f'Reopened by {user_displayname(user)}'

        self.notify_review_request(review_request,
                                   fallback_text=fallback_text,
                                   pre_text=pre_text,
                                   body=review_request.description,
                                   local_site=review_request.local_site,
                                   event_name='review_request_reopened')

    def _on_review_published(
        self,
        user: User,
        review: Review,
        **kwargs,
    ) -> None:
        """Handler for when a review is published.

        This will send a notification to any configured channels when
        a review is published.

        Args:
            user (django.contrib.auth.models.User):
                The user who published the review.

            review (reviewboard.reviews.models.Review):
                The review that was published.

            **kwargs (dict, unused):
                Additional keyword arguments passed to the handler.
        """
        fields: list[FieldsDict]
        open_issues = 0
        first_comment = None

        for comment in review.get_all_comments():
            if not first_comment:
                first_comment = comment

            if (comment.issue_opened and
                comment.issue_status == BaseComment.OPEN):
                open_issues += 1

        if open_issues == 1:
            issue_text = '1 issue'
        else:
            issue_text = f'{open_issues} issues'

        user_link = self.get_user_text_link(
            user=user,
            local_site=review.review_request.local_site)
        pre_text = f'New review from {user_link}'

        warning_emoji = '⚠'
        checkmark_emoji = '✅'

        if self.use_emoji_shortcode:
            warning_emoji = ':warning:'
            checkmark_emoji = ':white_check_mark:'

        if review.ship_it:
            if open_issues:
                fields = [{
                    'title': 'Fix it, then Ship it!',
                    'value': self.format_field_text(
                        f'{warning_emoji} {issue_text}'),
                    'short': True,
                }]
                extra_text = ' (Fix it, then Ship it!)'
                color = 'warning'
            else:
                fields = [{
                    'title': 'Ship it!',
                    'value': self.format_field_text(checkmark_emoji),
                    'short': True,
                }]
                extra_text = ' (Ship it!)'
                color = 'good'
        elif open_issues:
            fields = [{
                'title': 'Open Issues',
                'value': self.format_field_text(
                    f'{warning_emoji} {issue_text}'),
                'short': True,
            }]
            extra_text = f' ({issue_text})'
            color = 'warning'
        else:
            fields = []
            extra_text = ''
            color = None

        fallback_text = f'New review from {user_displayname(user)}{extra_text}'

        self.notify_review_or_reply(user=user,
                                    review=review,
                                    pre_text=pre_text,
                                    fallback_text=fallback_text,
                                    first_comment=first_comment,
                                    fields=fields,
                                    color=color,
                                    event_name='review_published')

    def _on_reply_published(
        self,
        user: User,
        reply: Review,
        **kwargs,
    ) -> None:
        """Handler for when a reply to a review is published.

        This will send a notification to any configured channels when
        a reply to a review is published.

        Args:
            user (django.contrib.auth.models.User):
                The user who published the reply.

            reply (reviewboard.reviews.models.Review):
                The reply that was published.

            **kwargs (dict, unused):
                Additional keyword arguments passed to the handler.
        """
        user_link = self.get_user_text_link(
            user=user,
            local_site=reply.review_request.local_site)
        pre_text = f'New reply from {user_link}'
        fallback_text = f'New reply from {user_displayname(user)}'

        self.notify_review_or_reply(user=user,
                                    review=reply,
                                    fallback_text=fallback_text,
                                    pre_text=pre_text,
                                    event_name='reply_published')



"""Integration for Discord"""

from __future__ import annotations

import json
import logging
from typing import MutableMapping, Optional, Sequence, TYPE_CHECKING
from urllib.request import Request, urlopen

from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from rbintegrations.basechat.integration import BaseChatIntegration
from rbintegrations.discord.forms import DiscordIntegrationConfigForm
from rbintegrations.slack.integration import format_link, build_slack_message
from rbintegrations.util.compat.logs import log_timed

if TYPE_CHECKING:
    from djblets.util.typing import JSONDict
    from rbintegrations.basechat.integration import FieldsDict
    from reviewboard.reviews.models import ReviewRequest
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


class DiscordIntegration(BaseChatIntegration):
    """Integrates Review Board with Discord.

    This will handle updating Discord channels when review requests are
    posted, changed, or closed, and when there's new activity on the review
    request.
    """

    name = 'Discord'
    description = _(
        'Notifies channels in Discord when review requests are created, '
        'updated, and reviewed.'
    )

    default_settings = {
        'channel': '',
        'notify_username': 'Review Board',
        'webhook_url': '',
    }

    config_form_cls = DiscordIntegrationConfigForm

    assets_base_url = (
        'https://static.reviewboard.org/integration-assets/discord'
    )

    def notify(
        self,
        *,
        title: str,
        title_link: str,
        fallback_text: str,
        local_site: Optional[LocalSite],
        review_request: ReviewRequest,
        event_name: Optional[str] = None,
        fields: Sequence[FieldsDict] = [],
        pre_text: Optional[str] = None,
        body: Optional[str] = None,
        color: Optional[str] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Send a WebHook notification to Discord.

        This will post the given message to any Discord channels configured to
        receive it.

        Args:
            title (str):
                The title for the message.

            title_link (str):
                The link for the title of the message.

            fallback_text (str):
                The non-rich fallback text to display in the chat, for use in
                IRC and other services.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site for the review request or review emitting
                the message. Only integration configurations matching this
                Local Site will be processed.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the notification is bound to.

            event_name (str, optional):
                The name of the event triggering this notification.

            fields (list of rbintegrations.basechat.integration.FieldsDict,
                    optional):
                The fields comprising the rich message to display in chat.

            pre_text (str, optional):
                Text to display before the rest of the message.

            body (str, optional):
                The body of the message.

            color (str, optional):
                A color string or RGB hex value for the message.

            thumb_url (str, optional):
                URL of an image to show on the side of the message.

            image_url (str, optional):
                URL of an image to show in the message.
        """
        common_payload = build_slack_message(integration=self,
                                             title=title,
                                             title_link=title_link,
                                             fallback_text=fallback_text,
                                             fields=fields,
                                             pre_text=None,
                                             body=body,
                                             color=color,
                                             thumb_url=thumb_url,
                                             image_url=image_url)
        common_payload['text'] = pre_text

        # Send a notification to any configured channels.
        for config in self.get_configs(local_site):
            if not config.match_conditions(form_cls=self.config_form_cls,
                                           review_request=review_request):
                continue

            payload: JSONDict = dict({
                'username': config.get('notify_username'),
            }, **common_payload)

            # Tell Discord to use Slack formatted message.
            webhook_url = '%s/slack' % config.get('webhook_url')

            with log_timed(f'Sending notification for event "{event_name}", '
                           f'review_request ID {review_request.pk}, '
                           f'WebHook URL {webhook_url}',
                           logger=logger) as log_timer:
                try:
                    data = json.dumps(payload).encode('utf-8')

                    # Must include 'User-Agent' to avoid being blocked by
                    # Discord.
                    headers: MutableMapping[str, str] = {
                        'Content-Length': str(len(data)),
                        'Content-Type': 'application/json',
                        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64)',
                    }

                    urlopen(Request(webhook_url, data, headers))
                except Exception as e:
                    error_id = log_timer.trace_id
                    fp = getattr(e, 'fp', None)

                    logger.exception('[%s] Failed to send notification: %s',
                                     error_id, e)
                    logger.debug('[%s] Discord message payload = %r',
                                 error_id, payload)

                    if fp is not None:
                        logger.debug('[%s] Discord error response = %r',
                                     error_id, fp.read())

    def format_link(
        self,
        *,
        path: str,
        text: str,
    ) -> str:
        """Format the given URL and text to be shown in a Slack message.

        This uses SlackIntegration's format_link function and Discord will
        handle the translation of Slack URL to Discord URL

        Args:
            path (str):
                The path on the Review Board server.

            text (str):
                The text for the link.

        Returns:
            str:
            The link for use in Slack.
        """
        return format_link(path=path, text=text)

    def format_field_text(
        self,
        text: str,
    ) -> str:
        """Format the field text, providing any normalization required.

        This will limit the text of any Discord field to 1024 characters.
        Failing to do so results in errors posting messages.

        Version Added:
            4.0

        Args:
            text (str):
                The text for the field.

        Returns:
            str:
            The formatted or normalized text.
        """
        if len(text) > 1024:
            text = '%s...' % text[:1021]

        return text

    # NOTE: It's technically type-unsafe to replace an attribute with a
    #       property, but this works for us practically.
    @cached_property
    def icon_static_urls(self) -> dict[str, str]:  # type: ignore
        """Return the icons used for the integration.

        Returns:
            dict:
            The icons for Discord.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        assert extension is not None

        return {
            '1x': extension.get_static_url('images/discord/icon.png'),
            '2x': extension.get_static_url('images/discord/icon@2x.png'),
        }



"""URL definitions for the Asana integration."""

from __future__ import annotations

from django.urls import include, path

from rbintegrations.asana.views import (AsanaTaskSearchView,
                                        AsanaWorkspaceListView)


localsite_urlpatterns = [
    path('task-search/<int:review_request_id>/',
         AsanaTaskSearchView.as_view(),
         name='asana-task-search'),
]

urlpatterns = [
    path('workspaces/', AsanaWorkspaceListView.as_view(),
         name='asana-workspace-list'),
    path('s/<local_site_name>/', include(localsite_urlpatterns)),
]
urlpatterns += localsite_urlpatterns



"""The form for configuring the Travis CI integration."""

from __future__ import annotations

import logging
from urllib.error import HTTPError, URLError
from typing import Iterator, TYPE_CHECKING

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.conditions.choices import ConditionChoices
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.reviews.conditions import (ReviewRequestConditionChoiceMixin,
                                            ReviewRequestConditionChoices,
                                            ReviewRequestRepositoriesChoice,
                                            ReviewRequestRepositoryTypeChoice)
from reviewboard.scmtools.conditions import RepositoriesChoice
from reviewboard.scmtools.models import Repository

from rbintegrations.baseci.forms import BaseCIIntegrationConfigForm
from rbintegrations.travisci.api import TravisAPI
from rbintegrations.util.compat.logs import log_timed
from rbintegrations.util.conditions import (ReviewRequestConditionsField,
                                            review_request_condition_choices)

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from djblets.conditions.choices import BaseConditionChoice


logger = logging.getLogger(__name__)


class GitHubRepositoriesChoice(ReviewRequestConditionChoiceMixin,
                               RepositoriesChoice):
    """A condition choice for matching a review request's repositories.

    This works the same as the built-in ``ReviewRequestRepositoriesChoice``,
    but limits the queryset to only be GitHub repositories.
    """

    def get_queryset(self) -> QuerySet[Repository]:
        """Return the queryset used to look up repository choices.

        This will limit the results to GitHub repositories.

        Returns:
            django.db.models.query.QuerySet:
            The queryset for repositories.
        """
        return (
            super().get_queryset()
            .filter(hosting_account__service_name='github')
        )

    def get_match_value(self, review_request, **kwargs):
        """Return the repository used for matching.

        Args:
            review_request (reviewboard.scmtools.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            reviewboard.scmtools.models.Repository:
            The review request's repository.
        """
        return review_request.repository


class GitHubOnlyConditionChoices(ConditionChoices):
    """A set of condition choices which limits to GitHub repositories.

    The basic ReviewRequestConditionChoices allows a little too much freedom to
    select repositories which can never work with Travis CI. This gets rid of
    the repository type condition, and limits the repository condition to only
    show GitHub repositories.
    """

    def get_defaults(self) -> Iterator[type[BaseConditionChoice]]:
        """Return defaults for the condition choices.

        This will generate defaults from the standard conditions for a
        review request, replacing the standard repository-related fields
        with one specific to Travis CI.

        Yields:
            djblets.conditions.choice.BaseConditionChoice:
            Each choice to include.
        """
        ignored = {
            ReviewRequestRepositoriesChoice,
            ReviewRequestRepositoryTypeChoice,
        }

        for choice in review_request_condition_choices:
            if choice not in ignored:
                yield choice

        yield GitHubRepositoriesChoice


class TravisCIIntegrationConfigForm(BaseCIIntegrationConfigForm):
    """Form for configuring Travis CI."""

    # Note that we're lazily constructing this so that we ensure that any
    # form instance will re-fetch the full list of conditions, ensuring
    # anything provided by extensions will be there and anything stale will
    # not.
    conditions = ReviewRequestConditionsField(
        choices=lambda: GitHubOnlyConditionChoices(),
        label=_('Conditions'))

    travis_endpoint = forms.ChoiceField(
        label=_('Travis CI'),
        choices=TravisAPI.ENDPOINT_CHOICES,
        help_text=_('The Travis CI endpoint for your project.'))

    travis_custom_endpoint = forms.URLField(
        label=_('CI Server'),
        required=False,
        help_text=_('The URL to your enterprise Travis CI server. For '
                    'example, <code>https://travis.example.com/</code>.'))

    travis_ci_token = forms.CharField(
        label=_('API Token'),
        help_text=(
            _('The Travis CI API token. To get an API token, follow the '
              'instructions at <a href="%(url)s">%(url)s</a>.')
            % {'url': 'https://developer.travis-ci.com/authentication'}))

    travis_yml = forms.CharField(
        label=_('Build Config'),
        help_text=_('The configuration needed to do a test build, without '
                    'any notification or deploy stages.'),
        widget=forms.Textarea(attrs={'cols': '80'}))

    branch_name = forms.CharField(
        label=_('Build Branch'),
        required=False,
        help_text=_('An optional branch name to use for review request '
                    'builds within the Travis CI user interface.'))

    def __init__(self, *args, **kwargs):
        """Initialize the form.

        Args:
            *args (tuple):
                Arguments for the form.

            **kwargs (dict):
                Keyword arguments for the form.
        """
        super(TravisCIIntegrationConfigForm, self).__init__(*args, **kwargs)

        from rbintegrations.extension import RBIntegrationsExtension
        extension = RBIntegrationsExtension.instance

        travis_integration_config_bundle = \
            extension.get_bundle_id('travis-ci-integration-config')
        self.css_bundle_names = [travis_integration_config_bundle]
        self.js_bundle_names = [travis_integration_config_bundle]

    def clean(self):
        """Clean the form.

        This validates that the configured settings are correct. It checks that
        the API token works, and uses Travis' lint API to validate the
        ``travis_yml`` field.

        Returns:
            dict:
            The cleaned data.
        """
        cleaned_data = super(TravisCIIntegrationConfigForm, self).clean()

        if self._errors:
            # If individual form field validation already failed, don't try to
            # do any of the below.
            return cleaned_data

        endpoint = cleaned_data['travis_endpoint']

        if (endpoint == TravisAPI.ENTERPRISE_ENDPOINT and
            not cleaned_data['travis_custom_endpoint']):
            self._errors['travis_custom_endpoint'] = self.error_class([
                _('The server URL is required when using an enterprise '
                  'Travis CI server.')
            ])
            return cleaned_data

        try:
            api = TravisAPI(cleaned_data)
        except ValueError as e:
            self._errors['travis_endpoint'] = self.error_class(
                [str(e)])
            return cleaned_data

        # First try fetching the "user" endpoint. We don't actually do anything
        # with the data returned by this, but it's a good check to see if the
        # API token is correct because it requires authentication.
        try:
            api.get_user()
        except HTTPError as e:
            if e.code == 403:
                message = _('Unable to authenticate with this API token.')
            else:
                message = str(e)

            self._errors['travis_ci_token'] = self.error_class([message])

            return cleaned_data
        except URLError as e:
            self._errors['travis_endpoint'] = self.error_class([e])
            return cleaned_data

        # Use the Travis API's "lint" endpoint to verify that the provided
        # config is valid.
        request = self.request

        with log_timed('Triggering Travis CI lint check',
                       logger=logger,
                       request=request) as log_timer:
            try:
                lint_results = api.lint(cleaned_data['travis_yml'])

                for warning in lint_results['warnings']:
                    if warning['key']:
                        if isinstance(warning['key'], list):
                            key = '.'.join(warning['key'])
                        else:
                            key = warning['key']

                        message = (_('In %s section: %s')
                                   % (key, warning['message']))
                    else:
                        message = warning['message']

                    self._errors['travis_yml'] = self.error_class([message])
            except URLError as e:
                logger.exception('[%s] Unexpected error when trying to lint '
                                 'Travis CI config: %s',
                                 log_timer.trace_id,
                                 e,
                                 extra={'request': request})
                self._errors['travis_endpoint'] = self.error_class([
                    _('Unable to communicate with Travis CI server.')
                ])
            except Exception as e:
                logger.exception('[%s] Unexpected error when trying to lint '
                                 'Travis CI config: %s',
                                 log_timer.trace_id,
                                 e,
                                 extra={'request': self.request})
                self._errors['travis_endpoint'] = self.error_class([e])

        return cleaned_data

    class Meta(BaseCIIntegrationConfigForm.Meta):
        fieldsets = (
            (_('What To Build'), {
                'description': _(
                    'You can choose which review requests to build using this '
                    'configuration by setting conditions here. At a minimum, '
                    'this should include the specific repository to use '
                    'this configuration for.'
                ),
                'fields': ('conditions',),
            }),
            (_('Where To Build'), {
                'description': _(
                    'Travis CI offers several different servers depending on '
                    'your project. Select that here and set up your API key '
                    'for the correct server.'
                ),
                'fields': ('travis_endpoint', 'travis_custom_endpoint',
                           'travis_ci_token'),
            }),
            (_('How To Build'), {
                'description': _(
                    "Builds performed on the code in review requests will use "
                    "a completely separate configuration from commits which "
                    "are pushed to the GitHub repository. The configuration "
                    "listed here will be used instead of the contents of the "
                    "repository's <code>.travis.yml</code> file. Note that "
                    "this should not contain any secret environment "
                    "variables."
                    "\n"
                    "It's also recommended to create a special branch head "
                    "in the GitHub repository to use for these builds, so "
                    "they don't appear to be happening on your main "
                    "development branch. This branch can contain anything "
                    "(or even be empty), since the code will come from the "
                    "review request."
                ),
                'fields': ('travis_yml', 'branch_name'),
                'classes': ('wide',)
            }),
            (_('When To Build'), {
                'fields': (
                    'run_manually',
                    'timeout',
                ),
            }),
        )



"""Forms for Discord Integration."""

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm

from rbintegrations.util.conditions import ReviewRequestConditionsField


class DiscordIntegrationConfigForm(IntegrationConfigForm):
    """Admin configuration form for Discord.

    This allows an administrator to set up a chat configuration for sending
    messages to a given Discord WebHook URL based on the specified
    conditions.
    """

    conditions = ReviewRequestConditionsField()

    webhook_url = forms.CharField(
        label=_('Webhook URL'),
        required=True,
        help_text=_('Your unique Discord webhook URL. This can be found at '
                    'the "Integration" page after selecting '
                    '"Server Settings".'),
        widget=forms.TextInput(attrs={
            'size': 80,
        }))

    class Meta:
        fieldsets = (
            ('What To Post', {
                'description': _(
                    'You can choose which review requests would be posted by '
                    'choosing the repositories and groups to match against.'
                ),
                'fields': ('conditions', ),
            }),
            ('Where To Post', {
                'description': _(
                    'To start, add a new webhook service integration on '
                    'Discord. You can then provide the unique webhook URL '
                    'below. Please ensure that the Discord webhook is tied '
                    'to the correct channel.'
                ),
                'fields': ('webhook_url', ),
                'classes': ('wide', )
            }),
        )



"""Integration for Mattermost"""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING

from django.utils.functional import cached_property

from rbintegrations.basechat.integration import BaseChatIntegration
from rbintegrations.slack.integration import format_link, notify

if TYPE_CHECKING:
    from rbintegrations.basechat.integration import FieldsDict
    from reviewboard.reviews.models import ReviewRequest
    from reviewboard.site.models import LocalSite


class MattermostIntegration(BaseChatIntegration):
    """Integrates Review Board with Mattermost.

    This will handle updating Mattermost channels when review requests are
    posted, changed, or closed, and when there's new activity on the review
    request.
    """

    name = 'Mattermost'
    description = (
        'Notifies channels in Mattermost when review requests are created, '
        'updated, and reviewed.'
    )

    default_settings = {
        'webhook_url': '',
        'channel': '',
        'notify_username': 'Review Board',
    }

    assets_base_url = (
        'https://static.reviewboard.org/integration-assets/mattermost'
    )

    def notify(
        self,
        *,
        title: str,
        title_link: str,
        fallback_text: str,
        local_site: Optional[LocalSite],
        review_request: ReviewRequest,
        event_name: Optional[str] = None,
        fields: Sequence[FieldsDict] = [],
        pre_text: Optional[str] = None,
        body: Optional[str] = None,
        color: Optional[str] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Send a WebHook notification to Mattermost.

        This will post the given message to any Mattermost channels
        configured to receive it.

        Args:
            title (str):
                The title for the message.

            title_link (str):
                The link for the title of the message.

            fallback_text (str):
                The non-rich fallback text to display in the chat, for use in
                IRC and other services.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site for the review request or review emitting
                the message. Only integration configurations matching this
                Local Site will be processed.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the notification is bound to.

            event_name (str, optional):
                The name of the event triggering this notification.

            fields (list of rbintegrations.basechat.integration.FieldsDict,
                    optional):
                The fields comprising the rich message to display in chat.

            pre_text (str, optional):
                Text to display before the rest of the message.

            body (str, optional):
                The body of the message.

            color (str, optional):
                A color string or RGB hex value for the message.

            thumb_url (str, optional):
                URL of an image to show on the side of the message.

            image_url (str, optional):
                URL of an image to show in the message.
        """
        notify(integration=self,
               title=title,
               title_link=title_link,
               fallback_text=fallback_text,
               local_site=local_site,
               review_request=review_request,
               event_name=event_name,
               fields=fields,
               pre_text=pre_text,
               body=body,
               color=color,
               thumb_url=thumb_url,
               image_url=image_url)

    def format_link(
        self,
        *,
        path: str,
        text: str,
    ) -> str:
        """Format the given URL and text to be shown in a Mattermost message.

        This will combine together the parts of the URL (method, domain, path)
        and format it using Mattermost's URL syntax.

        Args:
            path (str):
                The path on the Review Board server.

            text (str):
                The text for the link.

        Returns:
            str:
            The link for use in Slack.
        """
        return format_link(path=path, text=text)

    # NOTE: It's technically type-unsafe to replace an attribute with a
    #       property, but this works for us practically.
    @cached_property
    def icon_static_urls(self) -> dict[str, str]:  # type: ignore
        """The icons used for the integration.

        Returns:
            dict:
            The icons for Mattermost.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        assert extension is not None

        return {
            '1x': extension.get_static_url('images/mattermost/icon.png'),
            '2x': extension.get_static_url('images/mattermost/icon@2x.png'),
        }



"""Unit tests for Matrix"""

import json
from urllib.request import urlopen

from django.contrib.auth.models import User
from djblets.conditions import ConditionSet, Condition
from djblets.testing.decorators import add_fixtures
from reviewboard.accounts.trophies import TrophyType, trophies_registry
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import ReviewRequestDraft

from rbintegrations.matrix.integration import MatrixIntegration
from rbintegrations.testing.testcases import IntegrationTestCase


class MatrixIntegrationTests(IntegrationTestCase):
    """Tests for Matrix."""

    integration_cls = MatrixIntegration

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(MatrixIntegrationTests, self).setUp()

        self.user = User.objects.create(username='test',
                                        first_name='Test',
                                        last_name='User')

    def test_notify_new_review_request(self):
        """Testing MatrixIntegration notifies on new review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review request from Test User: '
                    'http://example.com/r/1/</strong><p>#1: Test Review '
                    'Request</p><strong><font '
                    'color="#efcc96">Description</font></strong><p>My '
                    'description.</p><strong><font '
                    'color="#efcc96">Repository</font></strong><p>Test '
                    'Repo</p><strong><font '
                    'color="#efcc96">Branch</font></strong><p>my-branch</p>',
                'format': 'org.matrix.custom.html'
            })

    @add_fixtures(['test_site'])
    def test_notify_new_review_request_with_local_site(self):
        """Testing MatrixIntegration notifies on new review request with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=False)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review request from Test User: '
                    'http://example.com/s/local-site-1/r/1/</strong><p>#1: '
                    'Test Review Request</p><strong><font color="#efcc96">'
                    'Description</font></strong><p>My '
                    'description.</p><strong><font color="#efcc96">'
                    'Repository</font></strong><p>Test Repo</p><strong><font '
                    'color="#efcc96">Branch</font></strong><p>my-branch</p>',
                'format': 'org.matrix.custom.html',
            })

    def test_notify_new_review_request_with_diff(self):
        """Testing MatrixIntegration notifies on new review request with diff
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)
        self.create_diffset(review_request)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review request from Test User: '
                    'http://example.com/r/1/</strong><p>#1: '
                    'Test Review Request</p><strong><font '
                    'color="#efcc96">Description</font></strong><p>My'
                    ' description.</p><strong><font '
                    'color="#efcc96">Diff</font></strong><p>http:'
                    '//example.com/r/1/diff/1/ | Revision '
                    '1</p><strong><font color="#efcc96">Repository'
                    '</font></strong><p>Test '
                    'Repo</p><strong><font color="#efcc96">Branch'
                    '</font></strong><p>my-branch</p>',
                'format': 'org.matrix.custom.html'
            })

    @add_fixtures(['test_site'])
    def test_notify_new_review_request_with_local_site_and_diff(self):
        """Testing MatrixIntegration notifies on new review request with local
        site and diff
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=False)

        self.create_diffset(review_request)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review request from Test User: '
                    'http://example.com/s/local-site-1/r/1/</strong>'
                    '<p>#1: Test Review Request</p><strong><font '
                    'color="#efcc96">Description</font></strong><p>My'
                    ' description.</p><strong><font '
                    'color="#efcc96">Diff</font></strong><p>'
                    'http://example.com/s/local-site-1/r/1/diff/1/ | Revision '
                    '1</p><strong><font color="#efcc96">Repository</font>'
                    '</strong><p>Test Repo</p><strong><font '
                    'color="#efcc96">Branch</font></strong><p>my-branch</p>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_request_with_fish_trophy(self):
        """Testing MatrixIntegration notifies on new review request with
        fish trophy
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)
        review_request.id = 12321
        review_request.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#12321: New review request from Test User: '
                    'http://example.com/r/12321/</strong>'
                    '<p>#12321: Test Review Request</p><strong><font '
                    'color="#efcc96">Description</font></strong>'
                    '<p>My description.</p><strong><font '
                    'color="#efcc96">Repository</font></strong>'
                    '<p>Test Repo</p><strong><font '
                    'color="#efcc96">Branch</font></strong>'
                    '<p>my-branch</p>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_request_with_milestone_trophy(self):
        """Testing MatrixIntegration notifies on new review request with
        milestone trophy
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            id=10000,
            publish=False)

        review_request.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body': (
                    '<strong>#10000: New review request from Test User: '
                    'http://example.com/r/10000/</strong><p>#10000: Test '
                    'Review Request</p><strong><font color="#efcc96">'
                    'Description</font></strong><p>My description.</p>'
                    '<strong><font color="#efcc96">Repository</font></strong>'
                    '<p>Test Repo</p><strong><font color="#efcc96">'
                    'Branch</font></strong><p>my-branch</p>'
                ),
                'format': 'org.matrix.custom.html',
            })

    def test_notify_new_review_request_with_custom_trophy(self):
        """Testing MatrixIntegration notifies on new review request with
        ignored custom trophy
        """
        class MyTrophy(TrophyType):
            category = 'test'

            def __init__(self):
                super(MyTrophy, self).__init__(title='My Trophy',
                                               image_url='blahblah')

            def qualifies(self, review_request):
                return True

        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)
        review_request.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)

        trophies_registry.register(MyTrophy)

        try:
            review_request.publish(self.user)
        finally:
            trophies_registry.unregister(MyTrophy)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review request from Test '
                    'User: http://example.com/r/1/</strong><p>#1: '
                    'Test Review Request</p><strong><font color="#efcc96">'
                    'Description</font></strong><p>My '
                    'description.</p><strong><font color="#efcc96">'
                    'Repository</font></strong><p>Test '
                    'Repo</p><strong><font color="#efcc96">Branch</font>'
                    '</strong><p>my-branch</p>',
                'format': 'org.matrix.custom.html'
            }
        )

    def test_notify_updated_review_request(self):
        """Testing MatrixIntegration notifies on updated review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New update from Test User: '
                    'http://example.com/r/1/</strong><p>#1: My new '
                    'summary</p><strong><font color="#efcc96">'
                    'Repository</font></strong><p>Test '
                    'Repo</p><strong><font color="#efcc96">'
                    'Branch</font></strong><p>my-branch</p>',
                'format': 'org.matrix.custom.html'
            }
        )

    def test_notify_updated_review_request_with_change_description(self):
        """Testing MatrixIntegration notifies on updated review request with
        change description
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        changedesc = draft.changedesc
        changedesc.text = 'These are my changes.'
        changedesc.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.publish(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New update from Test User: '
                    'http://example.com/r/1/</strong><p>#1: My new '
                    'summary</p><strong><font color="#efcc96">'
                    'Repository</font></strong><p>Test '
                    'Repo</p><strong><font '
                    'color="#efcc96">Branch</font></strong><p>'
                    'my-branch</p><blockquote>These are my '
                    'changes.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_closed_review_request_as_submitted(self):
        """Testing MatrixIntegration notifies on closing review request as
        submitted
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.close(review_request.SUBMITTED)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: Closed as completed by '
                    'Test User: http://example.com/r/1/</strong><p>#1: '
                    'Test Review Request</p>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_closed_review_request_as_discarded(self):
        """Testing MatrixIntegration notifies on closing review request as
        discarded
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.close(review_request.DISCARDED)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: Discarded by Test User:'
                    ' http://example.com/r/1/</strong><p>#1: Test '
                    'Review Request</p>',
                'format': 'org.matrix.custom.html'
            })

    @add_fixtures(['test_site'])
    def test_notify_closed_review_request_with_local_site(self):
        """Testing MatrixIntegration notifies on closing review request with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.close(review_request.SUBMITTED)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: Closed as completed by Test User: '
                    'http://example.com/s/local-site-1/r/1/'
                    '</strong><p>#1: Test Review Request</p>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_reopened_review_request(self):
        """Testing MatrixIntegration notifies on reopened review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)
        review_request.close(review_request.SUBMITTED)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.reopen(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: Reopened by Test User: '
                    'http://example.com/r/1/</strong><p>#1: Test Review '
                    'Request</p><blockquote>My description.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    @add_fixtures(['test_site'])
    def test_notify_reopened_review_request_with_local_site(self):
        """Testing MatrixIntegration notifies on reopened review request with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=True)
        review_request.close(review_request.SUBMITTED)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review_request.reopen(self.user)

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: Reopened by Test User: '
                    'http://example.com/s/local-site-1/r/1/'
                    '</strong><p>#1: Test Review '
                    'Request</p><blockquote>My description.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_with_body_top(self):
        """Testing MatrixIntegration notifies on new review with body_top"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='This is my review.')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User: '
                    'http://example.com/r/1/#review1</strong><p>#1: '
                    'Test Review Request</p><blockquote>'
                    'This is my review.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    @add_fixtures(['test_site'])
    def test_notify_new_review_with_local_site(self):
        """Testing MatrixIntegration notifies on new review with local site"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='This is my review.')
        self.create_general_comment(review)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User: '
                    'http://example.com/s/local-site'
                    '-1/r/1/#review1</strong><p>#1: Test Review '
                    'Request</p><blockquote>This is '
                    'my review.</blockquote>',
                'format': 'org.matrix.custom.html'
            }
        )

    def test_notify_new_review_with_comments(self):
        """Testing MatrixIntegration notifies on new review with only comments
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.')

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User: '
                    'http://example.com/r/1/#review1</strong><p>#1: '
                    'Test Review Request</p><blockquote>My '
                    'general comment.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_with_one_open_issue(self):
        """Testing MatrixIntegration notifies on new review with 1 open
        issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User (1 issue): '
                    'http://example.com/r/1/#review1</strong>'
                    '<p>#1: Test Review Request</p><strong><font '
                    'color="warning">Open Issues</font></strong>'
                    '<p>⚠ 1 issue</p><blockquote>My general '
                    'comment.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_with_open_issues(self):
        """Testing MatrixIntegration notifies on new review with > 1 open
        issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.',
                                    issue_opened=True)
        self.create_general_comment(review, text='My general comment 2.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User (2 issues): '
                    'http://example.com/r/1/#review1</strong>'
                    '<p>#1: Test Review Request</p><strong><font '
                    'color="warning">Open Issues</font></strong>'
                    '<p>⚠ 2 issues</p><blockquote>My general '
                    'comment.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_with_ship_it(self):
        """Testing MatrixIntegration notifies on new review with Ship It!"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User (Ship it!): '
                    'http://example.com/r/1/#review1</strong><p>'
                    '#1: Test Review Request</p><strong><font '
                    'color="good">Ship it!</font></strong>'
                    '<p>✅</p><blockquote>My comment</blockquote>',
                'format': 'org.matrix.custom.html'
            }
        )

    def test_notify_new_review_with_ship_it_and_custom_body_top(self):
        """Testing MatrixIntegration notifies on new review with Ship It and
        custom body_top
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='This is body_top.')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test User (Ship it!): '
                    'http://example.com/r/1/#review1'
                    '</strong><p>#1: Test Review Request</p><strong><font '
                    'color="good">Ship it!</font>'
                    '</strong><p>✅</p><blockquote>This is '
                    'body_top.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_with_ship_it_and_one_open_issue(self):
        """Testing MatrixIntegration notifies on new review with Ship It! and
        1 open issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review, issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test '
                    'User (Fix it, then Ship it!): '
                    'http://example.com/r/1/#review1'
                    '</strong><p>#1: Test Review Request</p><strong><font '
                    'color="warning">Fix it, then Ship it!'
                    '</font></strong><p>⚠ 1 issue</p>'
                    '<blockquote>My comment</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_review_with_ship_it_and_open_issues(self):
        """Testing MatrixIntegration notifies on new review with Ship It! and
        > 1 open issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review, issue_opened=True)
        self.create_general_comment(review, issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        review.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New review from Test '
                    'User (Fix it, then Ship it!): '
                    'http://example.com/r/1/#review1'
                    '</strong>'
                    '<p>#1: Test Review '
                    'Request</p><strong><font '
                    'color="warning">Fix it, then Ship '
                    'it!</font></strong><p>⚠ 2 issues</p>'
                    '<blockquote>My comment</blockquote>',
                'format': 'org.matrix.custom.html'
            }
        )

    def test_notify_new_reply_with_body_top(self):
        """Testing MatrixIntegration notifies on new reply with body_top"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review,
                                  user=self.user,
                                  body_top='This is body_top.')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        reply.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New reply from Test User: '
                    'http://example.com/r/1/#review1</strong><p>#1: '
                    'Test Review '
                    'Request</p><blockquote>This is body_top.'
                    '</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    @add_fixtures(['test_site'])
    def test_notify_new_reply_with_local_site(self):
        """Testing MatrixIntegration notifies on new reply with local site"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review,
                                  user=self.user,
                                  body_top='This is body_top.')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        reply.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New reply from Test User: '
                    'http://example.com/s/local-site-1/r/1/#review1'
                    '</strong><p>#1: Test Review '
                    'Request</p><blockquote>This is body_top.'
                    '</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def test_notify_new_reply_with_comment(self):
        """Testing MatrixIntegration notifies on new reply with comment"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review, user=self.user, body_top='')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)
        reply.publish()

        self.assertEqual(len(self.integration.notify.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'body': '',
                'msgtype': 'm.text',
                'formatted_body':
                    '<strong>#1: New reply from Test User: '
                    'http://example.com/r/1/#gcomment2</strong><p>#1: '
                    'Test Review Request</p><blockquote>This '
                    'is a comment.</blockquote>',
                'format': 'org.matrix.custom.html'
            })

    def _create_config(self, with_local_site=False):
        """Create an integration config.

        Args:
            with_local_site (bool, optional):
                Whether to limit the config to a local site.

        Returns:
            reviewboard.integrations.models.IntegrationConfig:
            A config for Matrix Integration to be used for testing.
        """
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any')),
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config 1',
                                                enabled=True,
                                                local_site=local_site)
        config.set('notify_username', 'RB User')
        config.set('access_token', 'matrix-example-token')
        config.set('room_id', 'matrix-example-id')
        config.set('conditions', condition_set.serialize())
        config.set('server', 'https://matrix.org')
        config.save()

        return config
